% Implemented by Henning Christiansen 1995 for the Demo system
%   
%
% Adapted for the new DemoII version of the system 1998 by
% Davide Martinenghi
%
% See legal notice in the end of this file.
%
% Load file as follows:
/*

      ['Examples/Fulladder'].

*/
%
% Diagnosis using the Demo system
% - developed by Henning Christiansen 1996

% The example is location of errors in logic circuits,
% the classical full adder example adapted from
% Example 2.2  p. 60 of
%
%    Raymond Reiter: A Theory of Diagnosis from First Principles
%    Artificial Intelligence 32, pp. 57-95, 1987


% We consider the following logical circuit in which
% each gate is identified by a label; it defines a full adder.
%
%                           CarryIn
%                              |
%                              |----))####\
%      A --------))####\       |    )) x2  >--- Sum
%          |     )) x1  >------(----))####/
%      B --(-----))####/   |   |
%          |   |           |   |
%          |   |           |   |
%          -----           -----
%          #   #           #   #
%          # a1#           # a2#
%          \___/           \___/
%            |               |
%            \               /
%             -----|   |-----
%                  |   |
%                  uuuuu
%                  #   #
%                  # o1#
%                  \___/
%                    |
%                    |
%                  CarryOut
%
% Components x1,x2 are exclusive-or gates, a1,a2 are 'and' gates,
% o1 is an 'or' gate.
%
% We can represent the circuit as an object program as follows;
% the internal 'wires' are represented by variables X, Y, and Z.

object_module( fulladder,

          \ [ (fulladder('A', 'B', 'CarryIn', 'Sum', 'CarryOut'):-

                   xorgate(x1, 'A',       'B',       'X'),
                   andgate(a1, 'A',       'B',       'Y'),
                   andgate(a2, 'X',       'CarryIn', 'Z'),
                   xorgate(x2, 'CarryIn', 'X',       'Sum'),
                   orgate( o1, 'Y',       'Z',       'CarryOut') )] ).

% The behaviour of the gates is described in the following object
% program; if a given component's status is 'ok' it behaves as
% states in a correct truth table, otherwise it is unpredictable.


object_module( gates,

         \ [ (andgate('Ident', 'In1', 'In2', 'Out'):-
                   status('Ident', ok),
                   and('In1', 'In2', 'Out')
              ),
              (andgate('Ident', 'In1', 'In2', 'Out'):- status('Ident', not_ok)
              ),

              (xorgate('Ident', 'In1', 'In2', 'Out'):-
                   status('Ident', ok),
                   xor('In1', 'In2', 'Out')
              ),
              (xorgate('Ident', 'In1', 'In2', 'Out'):- status('Ident', not_ok)
              ),

              (orgate('Ident', 'In1', 'In2', 'Out'):-
                   status('Ident', ok),
                   or('In1', 'In2', 'Out')
              ),
              (orgate('Ident', 'In1', 'In2', 'Out'):- status('Ident', not_ok)
              )
          ]).

object_module( truth_table,

         \ [  and(0, 0, 0),
              and(0, 1, 0),
              and(1, 0, 0),
              and(1, 1, 1),

              xor(0, 0, 0),
              xor(0, 1, 1),
              xor(1, 0, 1),
              xor(1, 1, 0),

              or(0, 0, 0),
              or(0, 1, 1),
              or(1, 0, 1),
              or(1, 1, 1)
          ]).

% If we believe that our circuit is perfect, we can state
% this by the following object program.

object_module( perfect_fa, \[

              status(x1, ok),
              status(x2, ok),
              status(a1, ok),
              status(a2, ok),
              status(o1, ok)  ]).

% We can calculate with the perfect fulladder in the following way:

/*

      constant_(S), constant_(C),
      demo( \ (perfect_fa & gates & fulladder & truth_table),
            \\fulladder(1,1,0,?S,?C)).

*/

% or the following to generate a truth table on backtracking:
/*

      constant_(A), constant_(B), constant_(Cin),
      constant_(S), constant_(Cout),
      demo( \ (perfect_fa & gates & fulladder & truth_table),
            \\fulladder(?A,?B,?Cin,?S,?Cout)).
*/

% The more interesting case is when incorrect behaviour
% is observed and the question is to identify the components
% in the circuit that, if they are the ones out of order, can explain
% the symptoms.
%
% The following metapredicate defines the general pattern
% - in terms of object programs -
% describing the state of the circuit, notice that it
% is parameterized by metavariables that stand for
% the actual status of each component.

status_for_full_adder(
   [OkOrNok_x1,OkOrNok_x2,OkOrNok_a1,OkOrNok_a2,OkOrNok_o1],

        \[

              status(x1, ?OkOrNok_x1),
              status(x2, ?OkOrNok_x2),
              status(a1, ?OkOrNok_a1),
              status(a2, ?OkOrNok_a2),
              status(o1, ?OkOrNok_o1)  ]):-

              constant_(OkOrNok_x1),
              constant_(OkOrNok_x2),
              constant_(OkOrNok_a1),
              constant_(OkOrNok_a2),
              constant_(OkOrNok_o1),

              ok_or_not_ok(OkOrNok_x1),
              ok_or_not_ok(OkOrNok_x2),
              ok_or_not_ok(OkOrNok_a1),
              ok_or_not_ok(OkOrNok_a2),
              ok_or_not_ok(OkOrNok_o1).

:- block ok_or_not_ok(-). % i.e. always delay until last moment.

ok_or_not_ok( \\\ok).
ok_or_not_ok( \\\not_ok).

% Example:
% Giving A=1, B=1, CarryIn = 0 to the fulladder will under
% normal circumstances produce Sum=0, CarryOut=1, but assume
% we observe the result Sum=1, CarryOut=1.
%
% At least one component in the circuit must be out of order,
% and the following query will provide suggestions for which
% ones:

/*

      status_for_full_adder(L,S),
      demo( \ (?S & gates & fulladder & truth_table),
            \\fulladder(1,1,0,1,1)).

*/

% The first answer indicates that if gate x2 is out of order, then
% the behaviour can be explained --- then follows a large number
% of solutions, many of which state explanation suggesting
% that x2 together with other gates are out of order,
%
% A *diagnosis* is defined as a minimal explanation,
% which in our case means that the suggest set of out-of-order
% components is minimal, i.e., if any of them is assumed to
% by in order, then the symptoms cannot be explained any longer..
%
% For example, out of the 37 answers produced by the query above,
% only two are minimal, 'x2 is out of order' and
% 'x1 is out of order'.
%
% An example an explanation which is not a diagnosis is
% 'everything out of order' - obviously this state can explain
% everything, but it is not very useful for the repairer...

%%%%%%%%%%%%%%%%%%

% Now we will set up a control structure which produce the list
% of minimal solutions.

% Our strategy is based on the following conjecture:
%
%   Due to the ordering of the two clauses for ok_or_not_ok,
%   the call (status_for_circuit(S), demo( ...)), above
%   always produces a minimal solution first.
%
% (We still need to prove it).

% So we will ask for a first solution, then set up a suitable
% dif(-,-) test to express that the next solution must not
% be an extension of the one found, and then we call demo again.

% We continue this way calling demo for one solution constrained
% in this way by all previously found solutions.

% If, for example, one solution found says {x1,a2} are defect (and the
% remaining ones ok), then the following dif-test is set up:
%
%     dif([OkOrNok_x1,OkOrNok_a2], [\\\not_ok,\\\not_ok] ).


% The following Prolog code will set up the number of required dif-tests
% from a list of solutions already found.
% A solution is represented as a list of 5 elements,
% [OkOrNok_x1,OkOrNok_x2,OkOrNok_a1,OkOrNok_a2,OkOrNok_o1] each being
% either \\\ok or \\\not_ok we define this test as follows:


% not_contained_in(new-solution, list-of-previous-solutions)

not_contained_in(_,[]):- !.

not_contained_in(Xs, [Ss | More]):-

   diff_where_previous_has_not_ok( Xs, Ss),
   not_contained_in(Xs, More).

diff_where_previous_has_not_ok( Xs, Ss):-
   collect_pattern( Xs, Ss, PXs, PSs),
   dif(PXs, PSs).

collect_pattern([],[],[],[]):-!.

collect_pattern( [X|Xs], [\\\not_ok|Ss], [X|PXs], [\\\not_ok|PSs]):-
      !, collect_pattern( Xs, Ss, PXs, PSs).

collect_pattern( [_|Xs], [\\\ok|Ss], PXs, PSs):-
      !, collect_pattern( Xs, Ss, PXs, PSs).

%%%%%%%%%%%%%%%%%%
%
% The appropriate control structure is the following:

find_all_minimal_solutions(Observations, All):-
  find_all_minimal_solutions(Observations, [], All).

find_all_minimal_solutions(Observations, Previous, All):-
    status_for_full_adder(L,S),
    not_contained_in(L, Previous),
    demo( \ (?S & gates & fulladder & truth_table), Observations),
    !,
    find_all_minimal_solutions(Observations, [L | Previous], All).

find_all_minimal_solutions(_,S,S).

% Testrun:
/*

      find_all_minimal_solutions(\\fulladder(1,1,0,1,1),  M).

*/

% This gives the following answer
% M = [[\\\not_ok,\\\ok,\\\ok,\\\ok,\\\ok],
%      [\\\ok,\\\not_ok,\\\ok,\\\ok,\\\ok]]

%%%%%%%%%%%%%
%% For more testing: the following object language facts
%% defines the complete, correct behaviour, so one can
%% try perturbing this...


/*

      find_all_minimal_solutions(
         \\ ( fulladder(0,0,0,0,0),
              fulladder(0,0,1,1,0),
              fulladder(0,1,0,1,0),
              fulladder(0,1,1,0,1),
              fulladder(1,0,0,1,0),
              fulladder(1,0,1,0,1),
              fulladder(1,1,0,0,1),
              fulladder(1,1,1,1,1) ), M).

*/

% M = [[\\\ok,\\\ok,\\\ok,\\\ok,\\\ok]]

%%%%%%%%%%%%%%%%%%%%%%%
%% Here is one where one should imagine that the only diagnosis
%% would be 'whole circuit is out of order'...

/*
      find_all_minimal_solutions(
         \\ (
              fulladder(0,0,0,0,0),
              fulladder(0,0,0,1,0),
              fulladder(0,0,0,0,1),
              fulladder(0,0,0,1,1),

              fulladder(0,0,1,0,0),
              fulladder(0,0,1,1,0),
              fulladder(0,0,1,0,1),
              fulladder(0,0,1,1,1),

              fulladder(0,1,0,0,0),
              fulladder(0,1,0,1,0),
              fulladder(0,1,0,0,1),
              fulladder(0,1,0,1,1),

              fulladder(0,1,1,0,0),
              fulladder(0,1,1,1,0),
              fulladder(0,1,1,0,1),
              fulladder(0,1,1,1,1),

              fulladder(1,0,0,0,0),
              fulladder(1,0,0,1,0),
              fulladder(1,0,0,0,1),
              fulladder(1,0,0,1,1),

              fulladder(1,0,1,0,0),
              fulladder(1,0,1,1,0),
              fulladder(1,0,1,0,1),
              fulladder(1,0,1,1,1),

              fulladder(1,1,0,0,0),
              fulladder(1,1,0,1,0),
              fulladder(1,1,0,0,1),
              fulladder(1,1,0,1,1),

              fulladder(1,1,1,0,0),
              fulladder(1,1,1,1,0),
              fulladder(1,1,1,0,1),
              fulladder(1,1,1,1,1)
              ), M).

*/

%% but we get the following:
%
%  M = [[\\\not_ok,\\\ok,\\\not_ok,\\\not_ok,\\\ok],
%
%       [\\\not_ok,\\\not_ok,\\\not_ok,\\\ok,\\\ok],
%
%       [\\\not_ok,\\\ok,\\\ok,\\\ok,\\\not_ok],
%
%       [\\\ok,\\\not_ok,\\\not_ok,\\\not_ok,\\\ok],
%
%       [\\\ok,\\\not_ok,\\\ok,\\\ok,\\\not_ok]]


%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
