% Implemented by Henning Christiansen 1995 for the Demo system
%   
%
% Adapted for the new DemoII version of the system 1998 by
% Davide Martinenghi
%
% See legal notice in the end of this file.
%
% Load file as follows:
/*
      ['Examples/linear_drinks'].
*/
%
% Abduction under linear-logic-like conditions
% ============================================
%
% We show an alternative demo predicate which imitates a very
% small subset of linear logic and we give a simple example
% of abduction in this context.
%
% We view here each clause a *resource* - in the sense that
% when a clause has been used once in a proof, it is consumed.
%
% - See in the bottom of this file for references to background
%   material about linear logic.
%
% This demo predicate is specified as follows.
%
% demoL(P, Q) iff exists sigma: P |- Q sigma
%                 and each clause of P used at most once
%                 in the proof of Q.

% Our example is concerned with findiung out how to get drunk (sic!).
% The facts that can be abduces stands for individual drinks which,
% of course, you only can drink once.
% We formalize this as a meta-predicate as follows...

:- block abducible_program(-).

abducible_program( \[]).

abducible_program( \[?C | ?P] ):-
    abducible_clause(C),
    abducible_program(P).

:- block abducible_clause(-).

abducible_clause( \ (drink(?X):- true) ):-
    constant_(X), available(X).

%:- block available(-).

available( \\\tuborg ).
available( \\\another_tuborg ).
available( \\\carlsberg ).
available( \\\stella_artois ).
available( \\\1664 ).
available( \\\one_more_stella ).
available( \\\pint_o_guinness ).


% The following sample query will find explanations for how
% a person who gets drunk from three drinks can become drunk.
%
/*
    abducible_program(How),
    demoL( \([(drunk:- drink('_'), drink('_'), drink('_'))] & ?How),
           \\drunk),
    close_constraints(How).
*/

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Here comes the demo-predicate:


demoL(P, Q):-
   no_duplicates(Proof),
   program_(P),
   instance_(Q, Q1),
   demoL1(P, Q1, Proof, []).

% demoL1( P, Q, Proof1, Proof2)
% - P, Q as usual
% - Proof1-Proof2 a difference list for the proof of Q,
%   where a proof is defined as a list of clauses used.

demoL1(_, \\true, Proof, Proof).

demoL1(P, Atom, [Clause | Proof1], Proof2):-
   member_(Clause, P),
   instance_(Clause, \ (?Atom :- ?Body1) ),
   demoL1(P, Body1, Proof1, Proof2).

demoL1(P, \\ (? A, ?B), Proof1, Proof3):-
   demoL1(P, A, Proof1, Proof2),
   demoL1(P, B, Proof2, Proof3).


:- block no_duplicates(-).

no_duplicates( []).

no_duplicates( [X | Xs]):-
    not_in_list(X, Xs),
    no_duplicates(Xs).

:- block not_in_list(?, -).

not_in_list(_, []).

not_in_list(X, [X1 | Xs]):-
   dif(X, X1),
   not_in_list(X, Xs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  BACKGROUND
%
%  The theoretical background for linear logic is given by:
%
%      Girard, J.Y., Linear logic.
%      Theoretical Computer Science 50, pp. 1-101, 1987.
%
%  A programming language called Lolli based on linear logic
%  is described in:
%
%      Hodas, J.S. & Miller, D. Logic programming in a fragment of
%      linear logic.
%      Journal of Information and Control, To appear 1994
%
%  Our little experiment here indicates that it should be possible
%  also to build complete demo predicate for richer languages such
%  as Lolli.
%
%  To do this, the naming relation (and thus the selection of syntax
%  constraints) have to be extended to cope with Lolli's rich
%  collection of operators.
%
%  Otherwise (we believe), the basic mechanism of instance constraints
%  can be re-used - only the demo predicate needs more control.
%
%  For those who might want to play with the idea, we recommend also:
%
%      Cervesato, I. Lollipops taste of Vanilla too.
%      in: A.Momigliani & M.Ornaghi, eds.
%      'Proof-Theoretical Extensions of Logic Programming',
%      CMU, Pittsburgh, PA 15231-3890, USA.
%      Proceedings of an ICLP-94 Post-Conference Workshop, 1994.

%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
