% Implemented by Henning Christiansen 1995 for the Demo system
%   
%
% Adapted for the new DemoII version of the system 1998 by
% Davide Martinenghi
%
% See legal notice in the end of this file.
%
% Load file as follows:
/*
   ['Examples/still_life'].
*/

%
% ABDUCTION WITH INTEGRITY CONSTRAINTS
% ====================================
%
% Produced by Henning Christiansen 1994
% - adapted for new version of demo 1996
% - adapted for new-new version 1998
%
%
% The example is concerned with the relation between
% utterance about a certain scene and some programS
% of simple facts defining the scene.
%
% - see references about the linguistic background
%   in the bottom of the file.
%
% For example, the sentence 'the_flower is on the_table'
% is valid for a scene whose things are described by
% the following program,
%
%      thing(the_flower), thing(the_vase),
%      thing(the_table), container(the_vase)
%
% and where the physical placement of these objects are described
% by the following.
%
%      in(the_flower, the_vase), on(the_vase, the_flower).
%
% The grammar defining sentences is given by the
% object_module 'grammar' below.
%
% The integrity constraints saying what is means for an object
% program to represent a series of things (some of which
% are especially containers) is defined by a predicate
%
%    things_program( ThingTheory ).
%
% The integrity constraint which defines what it means to
% be a SceneTheory (given some ThingTheory) is defined
% by the predicate
%
%    scene_description(ThingTheory, SceneTheory).
%
% The general relation between a given sentence S
% things theory T and scene theory Sc can be informally stated
% as follows
%
%    things_program(T) /\ scene_description(T, Sc)
%    /\  grammar & T & Sc |= sentence(S)
% 
% Sample 'things' and 'scene' programs are provided
% by the object modules 'things1' and 'scene1' below.
%
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

object_module( grammar,

           \  [ (sentence('S'):- simple('S')),
                (sentence('S'):- folded('S')),
                (simple(['X', is, on, 'Y']):-
                    thing('X'),
                    thing('Y'),
                    on('X','Y')   ),
                (simple(['X', is, in, 'Y']):-
                    thing('X'),
                    thing('Y'),
                    in('X','Y')   ),
                (folded(['X', is, 'PREP', 'Y']):-
                    simple(['X', is, '_', 'Z']),
                    simple(['Z', is, 'PREP', 'Y'])  )
               ]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

object_module( things1,

           \  [ thing(the_flower),
                thing(the_vase),
                thing(the_table),
                thing(the_duck),
                container(the_vase) ]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

object_module( scene1,

           \  [ on(the_vase, the_flower),
                in(the_duck, the_vase),
                in(the_table, the_vase) ]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% scene_description(ThingTheory, SceneTheory)
%
% A SceneTheory describes the physical contacts between
% some things defined in some ThingTheory.
%
% A SceneTheory consists of facts in(-,-) and on(-,-)
% such that
% - each thing is defined in the given ThingTheory
%   as a thing(_) and (if needed) a container(_),
% - for any x, maximum one in(x,..) or on(x,..),
%   i.e., the x is only placed once.
% - we cannot have in(x,x) or on(x,x)
%
% These conditions give a reasonable (but not perfect)
% approximation to what can be interpreted as a 'still life'
% in the real world.
%
% (It appears, when you have a look at the code below, that
% the conditions could have been expressed in a more satisfactory
% manner if demo had been able to handle negation.)

:- block scene_description(?, -).

scene_description(_, \ []).

scene_description(T, \ [?F | ?S]):-
    scene_fact(T, F),
    scene_description(T, S) ,
    F = \ (?_ ^ [?ThisObject | ?_]  :- true),
    not_placed(ThisObject, S).

scene_fact(T, \ (in(?A,?B) :- true)):-
    constant_(A),
    constant_(B),
    demo(T, \\ (thing(?A), container(?B))),
    dif(A,B).

scene_fact(T, \ (on(?A,?B) :- true)):-
    constant_(A),
    constant_(B),
    demo(T, \\ (thing(?A), thing(?B))),
    dif(A,B).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% things_program( ThingTheory ).
%
% A ThingTheory is a program which defines a number
% of thing(..) and container(..) facts.
% - if container(x) then also thing(x).

things_program( ThingTheory ):-
    thing_and_container_facts_only( ThingTheory ),
    any_container_is_also_a_thing( ThingTheory, ThingTheory).

%%%%%%%%%%

:- block thing_and_container_facts_only(-).

thing_and_container_facts_only( \[] ).

thing_and_container_facts_only( \[ (thing(?T):- true) | ?P]):-
    constant_(T),
    thing_and_container_facts_only( P).

thing_and_container_facts_only( \[ (container(?T):- true) | ?P]):-
    constant_(T),
    thing_and_container_facts_only( P).

%%%%%%%%%%%%%
%
% any_container_is_also_a_thing(whole-program, rest-program)

:- block any_container_is_also_a_thing(?,-).

any_container_is_also_a_thing(_, \[]).

any_container_is_also_a_thing(WhP, \[(thing(? _):- true) | ?P]):-
    any_container_is_also_a_thing( WhP, P).

any_container_is_also_a_thing(WhP, \[(container(? T):- true) | ?P]):-
    demo( WhP, \\ thing(?T)),
    any_container_is_also_a_thing( WhP, P).


%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% not_placed( thing, program)
% - program does not contain facts about
%   on(thing,...) or in(thing,...)

:- block not_placed(?, -).

not_placed( _, \ []).

not_placed( Thing, \[ ( ? _ ^ [?Thing1, ? _] :- true)  | ?S]):-
   dif(Thing, Thing1),
   not_placed( Thing, S).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% Sample queries:
%
% Try e.g.,
/*

     mk_scene(X), close_constraints(X).

*/
%
% NB: 10-jul-98 OK, BUT REPEATS A SOLUTION!!!
% where:
%

mk_scene(X):- scene_description( \things1, X),
       demo( \ (grammar & things1 & ?X),
             \\(sentence([the_flower, is, on, the_table]))).

%
% Try e.g.,
/*


      mk_scene_and_things(X,T),
      close_constraints((X,T)).

*/
%
% or
/*
CURRENTLY THERE ARE NO EXTRA "PREFERENCE" ARGUMENTS
... will probably not be added again!!!!!!!
      mk_scene_and_things(X,T),
      close_constraints((X,T),[],[the_boat]).
?????????????????????????????????????????????????????????


*/
%
% where
% 

mk_scene_and_things(X,TT):- 
       things_program( TT ),
       scene_description(TT, X),
       demo( \ ( grammar & ?TT & ?X),
            \\ (sentence([the_duck, is, in, the_water]))).

%
% Try e.g.,
/*

      mk_sentence(S).

*/
%
% where
% 

mk_sentence([A,B,C,D]):-
     constant_(A),
     constant_(B),
     constant_(C),
     constant_(D),
     demo( \ ( grammar & things1 & scene1),
           \\ (sentence( [?A,?B,?C,?D] )) ).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  BACKGROUND FOR THE EXAMPLE
%
%    The example is an English transliteration and simplification
%    of a study in the Danish language.
% 
%    The Danish version of the example is actually more interesting
%    than what has been shown here. In Danish we rarely say
%    (transliterated!) "x IS in/on y". We prefer to use a more
%    informative verb than "is". Typically, we distinguish
%    between "sidder" (Eng.: "sits"), "ligger" (Eng.: "lies"),
%    and "st\aa r" (Eng.: "stands").
%
%    In English one may say "The flower is on the table" which
%    may mean several things....
%    In Danish we will distinguish cases by saying, e.g.,
%
%     - "Blomsten ligger p bordet"
%         (Eng. translit: "The flower lies on the table")
%        meaning that it lies flat on the table; if a vase is
%        involved, it must have turned over
%
%    or
%
%    - "Blomsten st\aa r p\aa{} bordet" 
%         (Eng. translit: "The flower stands on the table")
%      indication that the flower must be in something which
%      has the ability to 'stand' on a table by itself, e.g.,
%      a vase.
%
%    A witty account on the use of these verbs in Danish is
%    given by
%
%    Erik Hansen: "Sidde, st\aa{}, ligge".
%    M\aa l og M\ae le no. 2, 1974, pp. 26--32.
%    
%    ((NB: see your Tex/Latex book for the correct shape of letters
%    indicated by \..   ))
%
%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
