%%%%%%% This file is part of the DemoII system produced
%%%%%%% 1997-1999 by Davide Martinenghi and Henning Christiansen
%%%%%%%
%%%%%%% Contact:    henning@ruc.dk
%%%%%%%
%%%%%%% See legal notice in the end of this file.
%%%%%%%
%%%%%%% The code has been developed and tested for SICStus Prolog 3.7
%%%%%%%
%%%%%%% Filename: ac.pl

constraints preds_/2, no_preds_/2.


%%% Declarative semantics:
%%%
%%%   preds_(X, Spec)
%%%
%%%  Spec is a list of predicate names without arity.
%%%  A predicate is either a Prolog atomic or the
%%%  term par(ameter).
%%%
%%%  X is either
%%%  - a term of type program, clause, or atom,
%%%  - or prolog atomic or an item of the form
%%%    par(<var-id>) where <var-id> is what
%%%    the extended syntax believes is a variable.
%%%
%%%  The constraint is satisfied if and only if
%%%  the predicate symbol of each clause head in X,
%%%  resp. the predicate symbolf of X, or X itself
%%%  matches an element of Spec.
%%%
%%%  The term par(<var-id>) is said to match par(ameter)
%%%  and nothing else; any other symbol matches only itself.
%%%
%%%  If the arguments are not of the specified form, the result is
%%%  not specified
%%%
%%%  The Spec should be ground when called. (not tested)
%%%      (impl. by renaming preds_ -> preds_con_ and
%%%         preds_(X,S):- var(S) -> errormessage, abort ; preds_con_(X,L).
%%%       and anagously for no_preds_).



%%%  no_preds(X, Spec) is as above except that no predicate
%%%  name in any head (etc.) in X matches an element i Spec.

%%%  Example: preds_(P, [p,q,par(ameter)])

%%%  Pragmatics: The preds_ and no_preds can be used for any purpose
%%%  by the user, but they are primarily introduced in for
%%%  making it possible for fail to discard branches.


preds_(_, []) <=> fail.

preds_(X, Spec1), preds_(X, Spec2) <=>

    findall(S, (member(S,Spec1), member(S,Spec2)), Spec12),
    preds_(X, Spec12).

preds_(X, Spec1), no_preds_(X, Spec2) <=>

    findall(S, (member(S,Spec1), \+ member(S,Spec2)), Spec12),
    preds_(X, Spec12).

no_preds_(X, Spec1), no_preds_(X, Spec2) <=>

    append(Spec1, Spec2, Spec12),
    remove_duplicates(Spec12, Spec12NoDups),
    no_preds_(X, Spec12NoDups).

preds_(\ [], _) <=> true.

preds_(\ [?C|?Cs], Spec) <=> preds_(C, Spec), preds_(Cs, Spec).

preds_(\ (?H:- ?_), Spec) <=> preds_(H, Spec).

preds_(gr(atom, par(_), _), Spec) <=> member(par(ameter), Spec).

% WARNING: This one instantiates when predicate is unique
% Drop it?

preds_(gr(atom, Pred, _), [Pred]) <=> true.

preds_(gr(atom, Pred, _), Spec) <=> 
    ground(Pred)
  |
    member(Pred, Spec).

%%

no_preds_(\ [], _) <=> true.

no_preds_(\ [?C|?Cs], Spec) <=>
    no_preds_(C, Spec), no_preds_(Cs, Spec).

no_preds_(\ (?H:- ?_), Spec) <=> no_preds_(H, Spec).

no_preds_(gr(atom, par(_), _), Spec) <=> \+ member(par(ameter), Spec).

no_preds_(gr(atom, Pred, _), Spec) <=> 
    ground(Pred)
  |
    \+ member(Pred, Spec).

%%% Propagate through instance constraints:

preds_(X, Spec), instance_(X,Y,_) ==> preds_(Y,Spec).

no_preds_(X, Spec), instance_(X,Y,_) ==> no_preds_(Y,Spec).

%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1999, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/demo.html
