%%%%%%% This file is part of the DemoII system produced
%%%%%%% 1997-1999 by Davide Martinenghi and Henning Christiansen
%%%%%%%
%%%%%%% Contact:    henning@ruc.dk
%%%%%%%
%%%%%%% See legal notice in the end of this file.
%%%%%%%
%%%%%%% The code has been developed and tested for SICStus Prolog 3.7
%%%%%%%
%%%%%%% Filename: df_uti.pl



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The following works on internal terms
%
% can_select_simple(Query, Selected, Rest)
%
% serves as a guard as well as a selection rule
%
% -- fails if no interesting selection can be made.
%
% It is assumed, that Query is different from \\true
% - disjunctions (;) are selected as if they were 'simple' atoms.
% ?????????????
%
% Experiment of 7 may 1999:
%
% A principle that seems relevant for *non*recursive programs
%
% Priority in selection as follows:
%    1) Equalities that can be executed right away
%    2) any dif satisfying ?=
%    3) find the left-most atom
%
% (No 2. implies that 'we' keep control over the dif's instead
%  of having CHR or Prolog to do it!!!!!!)
%
% And the following improvements are possible:
%    - when an atom A has been selected and is tested against
%      a given clause H:-B and none of the usual rules apply,
%      execute instance_((H:-B),(H1:-B1)) and add
%      H1=A and B to the query (B possibly in the end in order
%      to get a breadth-first effect)
%    - no. 3 above changes into leftmost 'ripe' atom
%      (whatever this means)
%
% ... these two suggestions overlap a bit
%

% NB: sometimes a \\true can show up inside a query
% and we remove it immediately:

can_select_simple(Q,S,R):-
	can_select_true(Q,S,R), !.

can_select_simple(Q,S,R):-
	can_select_ripe_equality(Q,S,R), !.

can_select_simple(Q,S,R):-
	can_select_ripe_dif(Q,S,R), !.

can_select_simple(Q,S,R):-
	can_select_atom(Q,S,R), !.

%%%%

can_select_true(Q,_,_):- var(Q), !, fail.

can_select_true(\\true,\\true,\\true).

can_select_true(\\ (?A, ?B), Selected, Rest):-
	can_select_true(A, Selected, RestA)
	-> conjunc(RestA, B, Rest)
	;
	can_select_true(B, Selected, RestB)
	-> conjunc(A, RestB, Rest).


%%%
can_select_ripe_equality(Q,_,_):- var(Q), !, fail.

can_select_ripe_equality(\\ (?A = ?B), \\(?A = ?B), \\true):-
    ?=(A,B)
    ;
    no_more_changes_implied_by_meta_level((A,B)).

can_select_ripe_equality(\\ (?A, ?B), Selected, Rest):-
    can_select_ripe_equality(A, Selected, RestA)
    -> conjunc(RestA, B, Rest)
    ;
    can_select_ripe_equality(B, Selected, RestB)
    -> conjunc(A, RestB, Rest).


%%%%%%%

can_select_ripe_dif(Q,_,_):- var(Q), !, fail.

can_select_ripe_dif(\\ dif(?A, ?B), \\dif(?A,?B), \\true):-
    ?=(A,B).

can_select_ripe_dif(\\ (?A, ?B), Selected, Rest):-
    can_select_ripe_dif(A, Selected, RestA)
    -> conjunc(RestA, B, Rest)
    ;
    can_select_ripe_dif(B, Selected, RestB)
    -> conjunc(A, RestB, Rest).

%%%%%%%%


% For the time being, we take atoms left-> right

can_select_atom(Var, Var, \\true):-
    var(Var), !,
    chr:find_constraint(Var, type(atom,Var) ,_).

     % I.e., normally a variable means "don't know"
     % but when its type is known as shown, we can anyhow
     % select it - and for atoms, it might be the case
     % the preds_ (or no_preds_) constraints on V
     % is able to discarded branches

can_select_atom(gr(atom, Pred, Args),
                  gr(atom, Pred, Args), \\true).

can_select_atom(\\ (?A , ?B), Selected, Rest):-
    can_select_atom(A, Selected, ARest)
    -> conjunc(ARest, B, Rest)
    ;
    can_select_atom(B, Selected, RestB)
    -> conjunc(A, RestB, Rest).

% disjunctions treated as simple ones:

can_select_simple(\\ (?A ; ?B), \\ (?A ; ?B), \\true).

% Possible extensions: we may select dif's or ='s for which
% the arguments are ?= independently of left-to right
% because they can reduce immediately without delaying
% things that prevents other stuff from being executed.




% The following as an active conjunction that reduces
% \\true's away if possible 

conjunc(True,A,A):- True == \\true, !.
conjunc(A,True,A):- True == \\true, !.
conjunc(A,B,\\ (?A, ?B)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Guard used in first rule for fails1_with_clause
%
% atoms_cannot_instance_(ExternalAtom, InternalAtom, Continuation)
%  - Makes a 'syntactic' check to see if
%    the execution of instance_(ExternalAtom, InternalAtom)
%    for "obvious" reasons will fail
%
% If it succeeds, it is known for sure that instance always must fail.
% When it fails, this means "can't tell". 
%
% It has two purposes: it can apply even when the arguments
% not sufficiently instantiated as to execute instance_, and
% it is much more efficient than execution instance_
%
% This predicate should be a compromise between:
%   - remove as many as possible of those cases where
%     it can be shown that instance_(ExternalAtom, InternalAtom)
%     always will fail
%
%   - so efficient that we can use it as a guard
%
% It works by comparing predicates (taking possible
% preds_/no_preds_ constraints into account) and
% function symbols.
%
% For a metavariable V in either ExternalAtom or InternalAtom,
% we take a look in the possible dif(V,t) and dif(t,V) that
% reside in Continuation or in the Prolog execution state.
% The last option can be relevant when the no_duplicates
% condition set up by demo provide useful information.
%
% We do not take a full analysis of the dif's - only those that
% are easy.

atoms_cannot_instance_(E,I,_):-
%%% CAN WE LOOK AT SOME DIFFING HERE?????
    (var(E) ; var(I)),
    !,
    incompatible_preds_(E,I).

atoms_cannot_instance_(gr(atom,P1,As1), gr(atom,P2,As2),C):-
   
 (  atom(P1), atom(P2), P1 \== P2 -> true
   ;
    arg_lists_cannot_instance_(As1, As2,C) ).


% the following two could call args_cannot_instance(?)

arg_lists_cannot_instance_(E,_,_):- var(E), !, fail.
arg_lists_cannot_instance_(_,I,_):- var(I), !, fail.

arg_lists_cannot_instance_([],[],_):- !, fail.

arg_lists_cannot_instance_([A1|A1s],[A2|A2s],C):-
    args_cannot_instance_(A1,A2,C) -> true
   ;
    arg_lists_cannot_instance_(A1s,A2s,C).

args_cannot_instance_(E,I,C):-
    var(E), % only if we know that instance_(E,E)
            % always is the case, we can say something
    !,
    ((chr:find_constraint(E, type(constant,E),_)
      ;
      chr:find_constraint(E, ground_(E),_))
     -> is_dif(E,I,C)
     ;
     fail).

args_cannot_instance_(gr(Etype,_,_), _,_):-
    Etype == variable, !, fail.

args_cannot_instance_(E,I,C):-
    var(I), !,
    is_dif(I,E,C). % case where E has object variables ignored
                  % because they cannot get into I

args_cannot_instance_(gr(T1,F1,A1s), gr(T2,F2,A2s),C):-
     nonvar(T1), nonvar(T2), T1 \== T2 -> true
    ;
     nonvar(F1), nonvar(F2), F1 \== F2 -> true
    ;
      arg_lists_cannot_instance_(A1s,A2s,C).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% is_dif(Var, Term,C)
%
% succeeds if there is a dif(Var, T) or dif(T, Var) where
% T identical to Term 

is_dif(Var, Term, Cont):-
    frozen(Var, DelayedPrologCalls),
    (term_in_a_dif_prolog_comma(Term, DelayedPrologCalls)
     -> true
     ;
     term_in_a_dif_object_comma(Var, Term, Cont)).

%%
% term_in_a_dif_prolog_comma(T,CommaStruct)

term_in_a_dif_prolog_comma(T,(A,B)):-
    term_in_a_dif_prolog_comma(T,A) -> true
    ;
    term_in_a_dif_prolog_comma(T,B).

term_in_a_dif_prolog_comma(T, prolog:dif__int(X,Y)):-
   T==X ; T ==Y.

%%
% term_in_a_dif_object_comma(Var,T,CommaStruct)

term_in_a_dif_object_comma(_,_, Var):-
   var(Var), !, fail.

term_in_a_dif_object_comma(Var,T, \\ (?A, ?B)):-
    term_in_a_dif_object_comma(Var,T,A) -> true
    ;
    term_in_a_dif_object_comma(Var,T,B).

term_in_a_dif_object_comma(Var,T, Item):-
   Item == \\dif(?Var,?T)
   ;
   Item == \\dif(?T,?Var).
 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  skolemize_internal_variables(Term)

skolemize_internal_variables(Term):-
    term_variables(Term, VarList),
    skolemize_internal_variable_list(VarList).
    
skolemize_internal_variable_list([]).

skolemize_internal_variable_list([V|Vs]):-

   (var(V) ->  % i.e. not already skolemized
   
      % bypass external variables:
      ( chr:find_constraint(V, meta_level_quantified(V), _) -> true
        ;
        chr:find_constraint(V, instance_(_,V,_),_) -> true
        ;
         % this is an internal variable
        new_skolem_constant(V))
     ; true),
   skolemize_internal_variable_list(Vs).


:- dynamic skolem_next/1.

skolem_next(0).

new_skolem_constant(C):-
   retract( skolem_next(N)),
   N1 is N+1,
   assert( skolem_next(N1)),
   number_chars(N,NChars),
   append("$skolem", NChars, SkolemNChars),
   atom_chars(C, SkolemNChars).
   




%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Auxiliaries 
%

%%%%%%%
% The following necessary due to a bug in Sicstus Prolog 3.7:
    
subsumes_chk1(X,Y):-
    mk_variant(X,XX), mk_variant(Y,YY), subsumes_chk(XX,YY).

%%%%%
% no_more_changes_implied_by_meta_level(InternalTerm)

no_more_changes_implied_by_meta_level(T):-
    term_variables(T,Vars),
    no_more_changes_implied_by_meta_level_vars(Vars).

no_more_changes_implied_by_meta_level_vars([]).

no_more_changes_implied_by_meta_level_vars([V|Vs]):-
    chr:find_constraint(V, meta_level_quantified(V), _) -> fail
    ;
    chr:find_constraint(V, instance_(_,V,_),_) -> fail
    ;
    no_more_changes_implied_by_meta_level_vars(Vs).



%%%%%%%%%%%%%%
% unifiable_independently_of_external_variables(A,B)
% succeeds if replacing external variables with skolem constants A and B are unifiable

unifiable_independently_of_external_variables(A,B) :-
	get_external_variables((A,B),XV),
	mk_variant((A,B,XV),(ACopy,BCopy,XVCopy)),
	skolemize(XVCopy),
	ACopy=BCopy.	%We simulate the unification after skolemization of external vars

get_external_variables(Term,XV) :-
	term_variables(Term,Vars),
	discard_internal_variables(Vars,XV).

discard_internal_variables([],[]).
discard_internal_variables([V|Vs],XVars):-
      ( chr:find_constraint(V, meta_level_quantified(V), _) -> XV=[V]
        ;
        chr:find_constraint(V, instance_(_,V,_),_) -> XV=[V]
        ;
         % this is an internal variable
        XV=[]),

	discard_internal_variables(Vs,XVs),
	append(XV,XVs,XVars).

skolemize([]).

skolemize([V|Vs]):-
   (var(V) -> new_skolem_constant(V)
    ; true),
   skolemize(Vs).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% only_dif_constraints(SetOfDifs)
%
% checks that SetOfDifs is a set of only dif constraints
only_dif_constraints(Difs) :- var(Difs),!,fail.

only_dif_constraints(\\ dif(?_, ?_)).

only_dif_constraints(\\ (?A, ?B)) :-
	only_dif_constraints(A),
	only_dif_constraints(B).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% select_dif(SetOfDifs,SelectedDif,Rest)
%
% We assume that SetOfDifs is a set of difs
% It selects a dif (SelectedDif) from SetOfDifs
% and puts the rest in Rest
select_dif(\\ dif(?A, ?B), \\dif(?A,?B)).

select_dif(\\ (?A, ?B), Selected):-
    select_dif(A, Selected)
    ;
    select_dif(B, Selected).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% mk_internal_variant(X,Y,V,W) -
%  - Y is a variant of X with new variables
%  - and keeping variables covered by meta_level_quantified
%  - instance_ constraints pending on internal variables
%    are also copied and 'installes', with mk_internal_variant
%    applied recursively to their 2nd and 3rd arguments.
% - V and W constitute the variable's substitution list
%   in the same way as used by internal_variant
%
% ... there are probably often a large number of
%     duplicates among the constraints found, and
%     in order to limit recursion, we maintain
%     dynamically a list of those constraints that
%     have been processed.
%     This list is open in the end so new ones can be added there.
%
% The one that does the work is:
%        mk_internal_variant(X,Y,Cs,V,W) -


mk_internal_variant(V,W,VList,Wlist):-
    mk_internal_variant(V,W,_,VList,Wlist).


mk_internal_variant(V,W,Cs,VList,Wlist):-
    var(V), !,
    (chr:find_constraint(V, meta_level_quantified(V), _)
     -> V=W
     ;
     locate_internal(V,W,Cs,VList,Wlist,VList,Wlist)
    ).

mk_internal_variant(T,T1,Cs,VList,Wlist):-
    T=..[F|TL],
    mk_internal_variant1(TL,TL1,Cs,VList,Wlist),
    T1=..[F|TL1].

mk_internal_variant1([],[],_,_,_).

mk_internal_variant1([A|As],[B|Bs],Cs,VList,Wlist):-
    mk_internal_variant(A,B,Cs,VList,Wlist),
    mk_internal_variant1(As,Bs,Cs,VList,Wlist).

% analogous to locate/2 but includes constraint stuff

locate_internal(V,W,Cs,VL,WL,TotalVList,TotalWlist):-
    var(VL), !,
    % this means that V has not been met before,
    % so first enter V in the replacement table
    VL=[V|_], WL=[W|_],
    % and process the possible instance constraints pending
    % on V
    chr:findall_constraints(V, _, CsForV),
    % CsForV is a list of elements (Constraint # <irrelevant>)
    install_variant_constraints(CsForV,Cs, TotalVList,TotalWlist).
        


locate_internal(V,W,Cs,[V1|VL],[W1|WL], TotalVList,TotalWlist):-
     V==V1 -> W=W1
   ;
     locate_internal(V,W,Cs,VL,WL, TotalVList,TotalWlist).

%%%%%%%%%%%%%%%%%%%%%
% install_variant_constraints(CsList,Cs, VL,WL)
% - CsList is a list of elements (Constraint # <irrelevant>)
%   or delayed Prolog calls.
% - VL,WL is variables' replacement list
% - Cs is the open list of already processed constraints
% For each constraint found in the list, add a copy
% of it with 2nd and 3rd arguments processed by
% 

install_variant_constraints([],_,_,_):- !.

install_variant_constraints([(C#_)|List], Cs, VL,WL):-
   % this is a CHR constraints
    !, locate_constraint(C, Cs, Cs, VL,WL),
    install_variant_constraints(List, Cs, VL,WL).

install_variant_constraints([PrologCall|List], Cs, VL,WL):-
    locate_constraint(PrologCall, Cs, Cs, VL,WL),
    install_variant_constraints(List, Cs, VL,WL).
%%% 

locate_constraint(C,Cs,AllCs,VL,WL):-
    var(Cs), !,
    % first time C is met!
    
    % Add it to list:
    
    Cs=[C|_],

    % and install a copy of it the constraint store:

    (C=instance_(Arg1,Arg2,Arg3) % (this '=' does not seem to start CHR)
      ->  mk_internal_variant(Arg2,Arg2Copy,AllCs,VL,WL),
          mk_internal_variant(Arg3,Arg3Copy,AllCs,VL,WL),
          insert_constraint(instance_(Arg1, Arg2Copy, Arg3Copy), _)
      
     ; 
     C = diff(Arg1,Arg2)
      ->  mk_internal_variant(Arg1,Arg1Copy,AllCs,VL,WL),
          mk_internal_variant(Arg2,Arg2Copy,AllCs,VL,WL),
          insert_constraint(diff(Arg1Copy, Arg2Copy), _)
     ; 
     C = preds_(Arg, Spec)
      -> mk_internal_variant(Arg,ArgCopy,AllCs,VL,WL),
         insert_constraint(preds_(ArgCopy, Spec), _)
     ;
     C = no_preds_(Arg, Spec)
      -> mk_internal_variant(Arg,ArgCopy,AllCs,VL,WL),
         insert_constraint(no_preds_(ArgCopy, Spec), _)
     ;
%%%%%%     write('Funny constraint met by copy machine: '), write(C), nl,
     true
    ).


locate_constraint(C,[CC|Cs],AllCs,VL,WL):-
    C==CC -> true
   ;
    locate_constraint(C,Cs,AllCs,VL,WL).


%%%%%%%%%%%%%%%%%%%%%%%%
%
%   mk_variant
% used here as a more efficient version of mk_internal_variant
% - does not recurse through constraints

mk_variant(X,Y):- !, mk_variant(X,Y,_,_).

mk_variant(V,W,VList,Wlist):-
    var(V), !, locate(V,W,VList,Wlist).

mk_variant(T,T1,VList,Wlist):-
    T=..[F|TL], mk_variant1(TL,TL1,VList,Wlist), T1=..[F|TL1].

mk_variant1([],[],_,_).

mk_variant1([A|As],[B|Bs],VList,Wlist):-
    mk_variant(A,B,VList,Wlist), mk_variant1(As,Bs,VList,Wlist).

locate(V,W,VL,WL):-
    var(VL), !, VL=[V|_], WL=[W|_].

locate(V,W,[V1|_],[W|_]):- V==V1, !.

locate(V,W,[_|VL],[_|WL]):- locate(V,W,VL,WL).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  remove_dead_constraints(Term)
%
%  - removes all instance constraints pending in those variables 
%    in Term that are not  meta_level_quantified.
%  - There is no check that these constraints could not
%    have affected variables elsewhere!!!!
%    

%TESTPRINT
%remove_dead_constraints(X):-
%    nl,print(remove_dead_constraints(X)),nl, fail.

remove_dead_constraints(Var):-
    var(Var), !,
    (chr:find_constraint(Var, meta_level_quantified(Var), _) -> true
     ;
      remove_dead_constraints_from_var(Var)
    ).

% just repeat until all done

remove_dead_constraints_from_var(Var):-
   chr:find_constraint(Var, instance_(_,Arg2,Arg3), Id) ->
        (
		ground(Arg2) -> true ;
		chr:remove_constraint(Id),
		remove_dead_constraints(Arg2),
	        remove_dead_constraints_subst(Arg3),
        	remove_dead_constraints_from_var(Var)
        )
   ;
   chr:find_constraint(Var, preds_(Arg,_), Id) ->
        (chr:remove_constraint(Id),
         remove_dead_constraints(Arg),
         remove_dead_constraints_from_var(Var)
        )
   ;
   chr:find_constraint(Var, no_preds_(Arg,_), Id) ->
        (chr:remove_constraint(Id),
         remove_dead_constraints(Arg),
         remove_dead_constraints_from_var(Var)
        )

   ;
     true.

remove_dead_constraints(gr(_,_,Args)):-
    remove_dead_constraints_list(Args).

remove_dead_constraints_list(Var):-
    var(Var), !,
    remove_dead_constraints_from_var(Var).

remove_dead_constraints_list([]).

remove_dead_constraints_list([Arg|Args]):-
    remove_dead_constraints(Arg),
    remove_dead_constraints_list(Args).

%TESTPRINT
%remove_dead_constraints_subst(X):-
%    nl,print(remove_dead_constraints_subst(X)),nl, fail.

remove_dead_constraints_subst(Var):-
    var(Var), !,
    remove_dead_constraints_from_var(Var).

remove_dead_constraints_subst([(_ -> T)|More]):-
    remove_dead_constraints(T),
    remove_dead_constraints_subst(More).

remove_dead_constraints_subst([]). % probably never occurs


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% incompatible_preds_(P1, P2)
%
% Each of P1 and P2 can be one of:
%   - a variable of type atom, clause, or program
%        (perhaps constrained by preds_ or no_preds_)
%   - a structure indicating the name of a clause, perhaps
%        with a variable in the position of its head
%   - a structure indicating an atom
% However
%   - both should not be variables of type program
%   - a case where one arg. is a non-var term of type program
%     is excluded for reasons of efficiency.
%     (and the constraint solver for fail takes another route
%     anyhow).
%
% Should only be applied inside fails as in this file!!!!!

%
% The predicate is a test, if it succeeds
% it means for sure that predicate symbols of P1 and P2 are or cannot be
% instantiated to a 'common' predicate.
% For a clause, we refer to the predicate in its head, for
% a variable of type program, those predicates indicated by its
% possible preds_ or no_preds constraints.
%
% If it fails, it either means that they are identical or that
% nothing can be said.


incompatible_preds_(P1, P2):-
    % Define for each Pi either a PlusSeti or a MinusSeti
    % 

    find_good_plus_or_minus_set(P1, PlusSet1, MinusSet1),
    find_good_plus_or_minus_set(P2, PlusSet2, MinusSet2),
    
    % four cases:
    
    (nonvar(PlusSet1) ->
        % plus1:
        (nonvar(PlusSet2) ->
        
           % plus1-plus2 => disjoint:
           
           \+ (member(S,PlusSet1), member(S,PlusSet2))
           
         ; % plus1-minus2:  all of PlusSet1 found in MinusSet2
           
           \+ (member(S,PlusSet1), \+ member(S,MinusSet2))

        )
      ; % minus1:

        (nonvar(PlusSet2) ->
           % minus1-plus2: all of PlusSet2 found in MinusSet1
           
           \+ (member(S,PlusSet2), \+ member(S,MinusSet1))
           
         ; % minus1-minus2
           fail  % the two sets are finite, so there is always another
                 % element which is not in any of them
        )
      ).


% aux. only meaningful as used in incompatible_preds_ as above

find_good_plus_or_minus_set(X, PlusSet, MinusSet):-
    var(X)
      -> ( chr:find_constraint(X,preds_(X,PlusSet),_)
           -> ( member(par(_),PlusSet) -> fail; true)  % par() match everything
           ;
           chr:find_constraint(X,no_preds_(X,MinusSetZZ),_)
           -> delete(MinusSetZZ, par(ameter), MinusSet)  
           ;
           fail % we have no knowledge about X, so exit
         )
      ;
      X = gr(atom, Pred, _)
      -> ( var(Pred) -> fail ; Pred = par(_) -> fail ; PlusSet = [Pred] )
      ;
      X = \ (?Head :- ?_)
      -> find_good_plus_or_minus_set(Head, PlusSet, MinusSet).

%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
