%%%%%%% This file is part of the DemoII system produced
%%%%%%% 1997-1999 by Davide Martinenghi and Henning Christiansen
%%%%%%%
%%%%%%% Contact:    henning@ruc.dk
%%%%%%%
%%%%%%% See legal notice in the end of this file.
%%%%%%%
%%%%%%% The code has been developed and tested for SICStus Prolog 3.7
%%%%%%%
%%%%%%% Filename: es.pl

:- module(es, [switch/1, direct_subtype/2, subtype/2, lct/3, oper/5, lift_to_primary_type/2, reserved_symbol/1]).

:- multifile translate/2.
:- multifile user:term_expansion/2.
:- multifile user:gr/3.
:- multifile user:portray/1.
:- multifile user:portray_message/2.

:- dynamic translate/2.
:- dynamic te/1.

:- use_module(library(chr)).
:- use_module(loader).

:- demo_use_module('uti').
:- use_module(library(lists)).
:- use_module(library(terms)).
:- demo_use_module('es_types').
:- demo_use_module('es_operators').

translate(easy_mk_bias_predicate(Pred_name, Pattern, Condition),X) :- !,
%	true.
	translate_macro(macro1(Pred_name, Pattern, Condition),X).

%%% For structures, we allow extended syntax inside ordinary prolog predicates
translate(X,Y) :- !,X=..[F|Args],
	(quotation_operator(F) -> length(Args,Arity),print_message(warning,demo(quotator(F,Arity))) ; true),
	translate_list(Args,TArgs),Y=..[F|TArgs].

%%% if we find a variable, we don't touch it
translate_list(X,X):-var(X),!.
translate_list([],[]).
translate_list([A|Args],[TA|TArgs]) :-
	safe_translate(A,TA),
	translate_list(Args,TArgs).

%%% Backtranslation of uninstantiated variables
translate_back(X,X) :- var(X),!.

%%% Backtranslation of type constraints
translate_back(type(T,X),TC) :-
	( nonvar(T) -> type_exists(T) ) -> append_names(T,'_',TU), TC =.. [TU,X] ;
	TC = type(T,X).

%%% Backtranslation of check constraints
translate_back(prolog_check(C,X),PC) :-
	nonvar(C) -> append_names(C,'_',CU), PC =.. [CU,X] ;
	PC = check(C,X).

%%% Backtranslation of fails1 constraints
translate_back(fails1(A,B,_),F1C) :-
	te(garbage=on) -> fail ;
	tb(fails1(A,B),F1C,no_quotes).

%%% Backtranslation of fails1_with_all_clauses constraints
translate_back(fails1_with_all_clauses(A,B,C,D,_),F1WAC) :-
	te(garbage=on) -> fail ;
	tb(fails1_with_all_clauses(A,B,C,D),F1WAC,no_quotes).

%%% Backtranslation of fails1_with_clause constraints
translate_back(fails1_with_clause(A,B,C,D,_),F1WC) :-
	te(garbage=on) -> fail ;
	tb(fails1_with_clause(A,B,C,D),F1WC,no_quotes).

%%% Backtranslation of delay_remove_dead_constraints constraints
translate_back(delay_remove_dead_constraints(_,_),DRDC) :-
	te(garbage=on) -> fail ;
	write('a constraint for garbage '),DRDC=collection.

%%% Backtranslation of terms
translate_back(X,Y) :- tb(X,Y,no_quotes).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% WARNING: we make the assumption that
%%% a variable is represented as '$VAR'("Var_name")
%%% when printed out. So the predicate var2 is the
%%% same as var plus this new case.
%%% In all tb/3 we use var2 instead of var
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
var2(X) :- (var(X); X='$VAR'(_) ).

tb(X,X,no_quotes):-var2(X),!.
tb(X,?X,quotes):-var2(X),!.

%% the type is unspecified: we copy the whole structure as it is
tb(gr(Type,Name,Args),TTerm,Q) :- 
	var2(Type), !, question_mark(Q,gr(Type,Name,Args),TTerm).

%% if there is parameterization we translate into notation2
tb(gr(Type,ParName,Args),TTerm,Q) :-
	subsumes_chk(par(_),ParName),!,ParName=par(Name),
	if_var_put_qm(Name,QName),
	( var2(Args), !, QArgs=..[?,Args]; tblist(Args,QArgs,quotes) ),
	Notation2= QName^QArgs,
	quotator(Type,Quotator),
	quote(Quotator, Q, Notation2, TTerm).

%% name and/or args are unspecified: we translate into notation2
tb(gr(Type,Name,Args),TTerm,Q) :-
	( var2(Name),! ; var2(Args),! ; (\+ is_list(Args)),! ), !,
	if_var_put_qm(Name,QName),
	( var2(Args), !, QArgs=..[?,Args]; tblist(Args,QArgs,quotes) ),
	Notation2= QName^QArgs,
	quotator(Type,Quotator),
	quote(Quotator, Q, Notation2, TTerm).

%%%%%%%%%%%%%%%%% new case: object module found
tb(gr(program,Name,Args),TTerm,Q) :-
	ground((Name,Args)),
	object_module_source(Module_name,gr(program,Name,Args)), !, %% module found
	quotator(program,Quotator),
	quote(Quotator, Q, Module_name, TTerm).
%%%%%%%%%%%%%%%%% end of new case

%% type, name and args are nonvar: we translate into notation1
tb(gr(Type,Name,Args),TTerm,Q) :- !,
	(
	quotator(Type,Quotator),

	%%% this is a weird failure case (functor/predicate with weird name)
	%%% and the case of ambiguity with operators of different types but same quotator
	((\+ atom(Name), Args \== []) ;
	 (oper(Name,_,OtherType,_,_),dif(Type,OtherType),quotator(OtherType,Quotator)))
	 ->
	 ( Q==quotes -> TQArgs=Args ; tblist(Args,TQArgs,no_quotes) ),
	 question_mark(Q,gr(Type,Name,TQArgs),TTerm) ;

	tblist(Args,TArgs,quotes),
	(length(Args,0),!,Notation1=Name ; Notation1=..[Name|TArgs]),
	quotator(Type,Quotator), quote(Quotator, Q, Notation1, TTerm)
	).


%% Compound terms are not understood as a failure case
tb(Term,TTerm,no_quotes) :- compound(Term),!,Term=..[N|AL],
	translate_back_list(AL,TAL),TTerm=..[N|TAL].

translate_back_list([],[]):-!.
translate_back_list([X|TX],[Y|TY]) :- translate_back(X,Y), translate_back_list(TX,TY).

%% Atoms are translated into themselves
tb(Term,Term,_).

tblist(X,QX,Q) :- var2(X),!,question_mark(Q,X,QX).
tblist([],[],_):-!.
tblist([X|TX],[Y|TY],Q) :- tb(X,Y,Q), tblist(TX,TY,Q).

question_mark(quotes,Term,QTerm):- QTerm=..[?,Term].
question_mark(no_quotes,Term,Term).

if_var_put_qm(Var,QVar) :- 
	var2(Var), !, question_mark(quotes,Var,QVar) ;
	QVar=Var.

quote(Symbol,no_quotes,Term,QTerm):- QTerm=..[Symbol,Term].
quote(_,quotes,Term,Term).


switch([]).
switch([H|T]) :- switch(H), switch(T).
switch(X=Y) :- ( X==ahead; X==in; X==back; X==trace; X==error; X==garbage ),
	( Y==on; Y==off),
	( retract(te(X=_)),! ; true),
	assert(te(X=Y)).

:- switch([ahead=on, in=off, back=on, trace=off, error=on, garbage=off]).

user:portray(X) :- te(back=on),translate_back(X,Y),
	( Y=='$do_not_print' -> true ;
	writeq(Y) ).

user:portray_message(error,demo(E,T)) :- print_error_term(E,T).
user:portray_message(error,demo(E)) :- print_error(E).
user:portray_message(warning,demo(W)) :- print_warning(W).

print_error_term(Err,Term) :- error('{DEMO ERROR: '), error_type(Err), error('}'),
	errornl,error(Term),error(.),errornl.
print_error(Err) :- error('{DEMO ERROR: '), error_type(Err), error('}'), errornl.
print_warning(W) :- error('{DEMO WARNING: '), warning_type(W), error('}'), errornl.

error_type(expression(Expr)) :- error('wrong expression: '),errorq(Expr).
error_type(operator(O)) :- error('incorrect operator definition: '),error(O).
error_type(module) :- error('incorrect module definition').
error_type(macro) :- error('incorrect macro definition').
error_type(unexpected_variable) :- error('unexpected Prolog variable').
error_type(subtype(S)) :- error('illegal subtype definition: '''), error(S), error(''' is not a defined type').
error_type(merge(File)) :- error('missing constraint definition in '),error(File).

warning_type(singleton(F,S,L)) :- error('singleton variable '), errorq(S), error(' in file '),
	errorq(F), error(' - lines '), flatten(L,FL), min_list(FL,Min), max_list(FL,Max),
	error(Min),error(-),error(Max).

warning_type(quotator(Q,N)) :- error('quotation operator '), errorq(Q),
	error(' used as predicate name with arity '), errorq(N).

warning_type(close_constraints(X)) :- error('close_constraints could not complete its job - argument '),
	errorprint(X).

safe_translate(X,X) :- var(X),!.
safe_translate(X,Y) :- translate(X,Y).

no_error_translate(Old,New) :- on_exception(demo(_),safe_translate(Old,New),fail).

sound_translate(Old,New) :-
	on_exception(demo(E),safe_translate(Old,New),raise_exception(demo(E,Old))).

%%% Terms are manipulated by translate/2
%%% ?- represents queries
user:term_expansion( ?-(Old), ?-(New)) :- !,
	(
	var(Old) -> fail ;
	te(ahead=on) ->
		( safe_translate(Old,New), ! ; New=Old),
		( te(in=on) ->
			write('The internal form is: '), writeq(New), nl ;
			true
		) ;
	New=Old,
	(
	te(in=on) ->
		copy_term(Old,Fictitious),te(error=ERR),switch(error=off),
		(no_error_translate(Fictitious,Writable),! ; Writable='an error'),
		switch(error=ERR), write('The internal form would be: '),
		writeq(Writable),nl
	)
	).
	
user:term_expansion(Old, New) :- te(ahead=on) -> sound_translate(Old,New) ; New=Old.

%%% We are now ready to consult the metalanguage file containing
%%% the definitions for types and operators

:- demo_consult('type_definitions').
:- demo_use_module('es_macros').

:- write('{The extended syntax is ready}'),nl.

%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
