%%%%%%% This file is part of the DemoII system produced
%%%%%%% 1997-1999 by Davide Martinenghi and Henning Christiansen
%%%%%%%
%%%%%%% Contact:    henning@ruc.dk
%%%%%%%
%%%%%%% See legal notice in the end of this file.
%%%%%%%
%%%%%%% The code has been developed and tested for SICStus Prolog 3.7
%%%%%%%
%%%%%%% Filename: es_types.pl

:- module(es_types, [direct_subtype/2, subtype/2, type_exist/1, type_exists/1,
	quotation_operator/1, quotator/2, lct/3, lift_to_primary_type/2]).
:- use_module(library(lists)).
:- use_module(library(terms)).
:- use_module(loader).

:- demo_use_module('uti', [append_names/3]).

:- multifile es:translate/2.
:- dynamic quotator/2.
:- dynamic quotation_operator/1.
:- dynamic direct_subtype/2.

:- op(4, fy, [ type ] ).
:- op(750, fx, [ subtype ] ).
:- op(3, xfx, [ quoted ] ).
:- op(2, fx, [ by ] ).

/*****************************************************************
	Type definitions cause the following actions:
	1) The quotation operator is defined and stored;
	2) Inheritance from the supertype top is stored
	3) translate/2 has a clause handling the new type
*****************************************************************/
%%% Translation of type definitions
es:translate(type T quoted by Q, ( :- true ) ) :- !,
	( \+ quotator(_,Q) -> assert(quotation_operator(Q)) ; true),
	assert( quotator(T,Q) ),
	assert( direct_subtype(top,T) ),
	make_type_constraint(T),
	make_type_checker(T),
	op(3,fx, [ Q ] ),
%	consider_errors(Q),
	enrich_translate(Q).


/*****************************************************************
	A subtype definition has the same behavior as a type
	definition, except for the additional check that the supertype
	has already been defined and no error handling. The quotation
	operator is inherited from the supertype.
*****************************************************************/
%%% Translation of subtype definitions
es:translate(subtype S < T, ( :- true ) ) :- 
	\+ type_exists(T) ->
		raise_exception(demo(subtype(T))), !
	;
	!, assert( direct_subtype(T,S) ),
	quotator(T,Q),
	assert( quotator(S,Q) ),
	make_type_constraint(S),
	make_type_checker(S).

%%% asserts a clause for the type constraint
make_type_constraint(T) :-
	append_names(T,'_',TU),
	Head =.. [TU, X],
	assert( ( user:Head :- user:type(T,X) ) ).

%%% asserts a clause for the type checker
make_type_checker(T) :-
	append_names('is_',T,Is_T), append_names(Is_T,'_',Is_T_),
	Head =.. [Is_T_, X],
	assert( ( user:Head :- subsumes_chk(gr(T,_,_),X) ) ).

%%% Enriches es:translate/2 with a rule for type T quoted by Q
enrich_translate(Q) :- Arg=..[Q,X],
	asserta( (
		es:translate(Arg, TX) :- compatible(Q,TL),
		es_operators:sugar_translate(TL,_,X,TX),!
		) ).

%%% Check the existence of the type T
type_exists(top).
type_exists(T) :- direct_subtype(_,T).

%%% Check the existence of the types in the list
type_exist([]).
type_exist([H|T]) :- type_exists(H),type_exist(T).

%%% Retrieves the list of types compatible with Q
compatible(Q,TL) :- comp(Q,TL), \+ comp(Q,[_|TL]),!.
comp(_,[]).
comp(Q,[X|L]) :- quotator(X,Q),non_member(X,L),comp(Q,L).

%% Lift a type to its primary type
lift_to_primary_type([],[]) :- !.
lift_to_primary_type([H|T],[PH|PT]) :- !, lift_to_primary_type(H,PH),
	lift_to_primary_type(T,PT).
lift_to_primary_type(Special,PT) :- compound(Special), !, Special =.. [Name, Sym],
	lift_to_primary_type(Name,PName), PT =.. [PName, Sym].
lift_to_primary_type(S,PT) :- subtype(PT,S), direct_subtype(top,PT), ! ; PT = S.

%% transitive closure of direct_subtype
%subtype(T,T).
subtype(X,Y):-direct_subtype(X,Y).
subtype(X,Z):-direct_subtype(X,Y),subtype(Y,Z).
subtype(+(T),+(S)):-subtype(T,S).
subtype(*(T),*(S)):-subtype(T,S).
subtype(*(T),+(T)):-type_exists(T).	% +(T) is a subtype of *(T)
subtype(*(T),+(S)):-subtype(T,S).	% +(S) is a subtype of *(T)

subtype(*(T),TL):- subtype_list(TL,T).
subtype(+(T),[S|TL]):- subtype2(T,S),subtype_list(TL,T).
subtype([],[]).
subtype([T|TT],[S|ST]):-subtype2(T,S),subtype(TT,ST).

%subtype(_,bottom).
subtype2(T,T).
subtype2(T,S):-subtype(T,S).

subtype_list([],_).
subtype_list([S|TT],T):-subtype2(T,S),subtype_list(TT,T).

%% list version of direct_subtype
direct_subtypes(T,SL) :- d_subtypes(T,SL), \+ d_subtypes(T,[_|SL]),!.
d_subtypes(_,[]).
d_subtypes(T,[X|L]) :- direct_subtype(T,X),non_member(X,L),d_subtypes(T,L).


%% least common type
lct(T,T,T):-!.
lct(T1,T2,T1):-subtype(T2,T1),!.
lct(T1,T2,T2):-subtype(T1,T2),!.
lct(_,_,bottom):-fail.

%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
