%%%%%%% This file is part of the DemoII system produced
%%%%%%% 1997-1999 by Davide Martinenghi and Henning Christiansen
%%%%%%%
%%%%%%% Contact:    henning@ruc.dk
%%%%%%%
%%%%%%% See legal notice in the end of this file.
%%%%%%%
%%%%%%% The code has been developed and tested for SICStus Prolog 3.7
%%%%%%%
%%%%%%% Filename: uti.pl

:- module(uti,
	[trace/1, traceprint/1, tracenl/0, error/1, errornl/0, errorq/1, errorprint/1,

	product2list/2, list2comma/2, comma2list/2, flatten/2,

	is_variable/1, big_or_underscore/1, upper_case/1, lower_case/1,
	underscore/1, digit/1, first_letter/2, non_variable/1,

	append_names/3, fresh_list/2,

	merge/2,
	
	x/2]).

:- use_module(library(lists)).
:- use_module(library(terms)).
:- use_module(library(system)).

%%%%%%%%%%
% OUTPUT %
%%%%%%%%%%
tracenl :- es:te(trace=on) -> nl ; true.
trace(X) :- es:te(trace=on) -> write(X) ; true.
traceprint(X) :- es:te(trace=on) -> print(X) ; true.

errornl :- es:te(error=on) -> nl ; true.
error(X) :- es:te(error=on) -> write(X) ; true.
errorq(X) :- es:te(error=on) -> writeq(X) ; true.
errorprint(X) :- es:te(error=on) -> print(X) ; true.

%%%%%%%%%%%%%%%
% CONVERSIONS %
%%%%%%%%%%%%%%%

%%% Converts a product into a list
product2list(*(HP,T),L) :- !,product2list(HP,HL),append(HL,[T],L).
product2list(P,[P]).

%%% Converts a list into a comma structure
list2comma([],( true ) ) :- !.
list2comma([X],X):-!.
list2comma(.(H,LT),','(H,CT)):- list2comma(LT,CT).

%%% Converts a comma structure into a list
comma2list(','(H,CT),.(H,LT)):- comma2list(CT,LT).
comma2list(X,[X]).

%%%%%%%%%%%%%%%%%%%%%
% NAMES AND STRINGS %
%%%%%%%%%%%%%%%%%%%%%

%%% Checks if the argument looks like a Prolog variable
is_variable(V) :- 
	atomic(V), name(V,[H|T]),
	big_or_underscore(H),
	well_formed(T).

non_variable(V) :- 
	first_letter(V,H),
	big_or_underscore(H),!,fail.
non_variable(_).

big_or_underscore(H) :-
	upper_case(H) ;
	underscore(H).

upper_case(C) :- C >= 65, C =< 90.
lower_case(C) :- C >= 97, C =< 122.
underscore(C) :- C == 95.
digit(C) :- C >= 48, C =< 57.

first_letter(V,F) :- atomic(V), name(V,[F|_]),!.
first_letter(V,F) :- V =.. [Head|_], name(Head,[F|_]).

well_formed([]) .
well_formed([H|T]) :-
	( upper_case(H) ;
	lower_case(H) ;
	underscore(H) ;
    digit(H) ) ,
	well_formed(T).

%%% Appends two names
append_names(A,B,AB) :- name(A,NA), name(B,NB), append(NA,NB,NAB), name(AB,NAB).

%%% generates a list of N fresh variables
fresh_list(0,[]) :- !.
fresh_list(N,[_|T]) :- M is N-1,fresh_list(M,T).


%%%%%%%%%%%%%%%%
% File merging %
%%%%%%%%%%%%%%%%

merge([File],Dest) :- put_handler(File, Dest).
merge(Files,Dest) :- merge_files(Files, demo_temp), put_handler(demo_temp, Dest),
	delete_file(demo_temp).

merge_files([F1,F2],Merged) :- !, warn_singleton(F1),warn_singleton(F2),merge_constraints(F1,F2,Merged).
merge_files([F|Files],Merged) :- length(Files,N),N>=2,!, warn_singleton(F), append_names(Merged,N,MergedN),
	merge_files(Files,MergedN), merge_constraints(F,MergedN,Merged),delete_file(MergedN).

merge_constraints(File1, File2, Merged) :-
	seeing(OldInput),
	telling(OldOutput),
	tell(Merged),


	see(File1),
	% writes 'use_module' clauses from the first file (if any)
	repeat,
		read_term(Term1,[syntax_errors(dec10)]),
		((subsumes_chk((:- use_module(_)),Term1);subsumes_chk((:- use_module(_,_)),Term1);
		subsumes_chk((:- demo_use_module(_)),Term1);subsumes_chk((:- demo_use_module(_,_)),Term1))
			-> write_file(Term1) ; true),
		\+ (subsumes_chk((:- use_module(_)),Term1);subsumes_chk((:- use_module(_,_)),Term1);
		subsumes_chk((:- demo_use_module(_)),Term1);subsumes_chk((:- demo_use_module(_,_)),Term1)),
		!,

	see(File2),
	% writes 'use_module' clauses from the second file (if any)
	repeat,
		read_term(Term2,[syntax_errors(dec10)]),
		((subsumes_chk((:- use_module(_)),Term2);subsumes_chk((:- use_module(_,_)),Term2);
		subsumes_chk((:- demo_use_module(_)),Term2);subsumes_chk((:- demo_use_module(_,_)),Term2))
			-> write_file(Term2) ; true),
		\+ (subsumes_chk((:- use_module(_)),Term2);subsumes_chk((:- use_module(_,_)),Term2);
		subsumes_chk((:- demo_use_module(_)),Term2);subsumes_chk((:- demo_use_module(_,_)),Term2)),
		!,


	( Term1 = constraints(C1) -> comma2list(C1,L1) ;
	told, seen, see(OldInput), tell(OldOutput), raise_exception(demo(merge(File1))) ),
	( Term2 = constraints(C2) -> comma2list(C2,L2) ;
	told, seen, see(OldInput), tell(OldOutput), raise_exception(demo(merge(File2))) ),
	append(L1,L2,L3), list2comma(L3,C3), Term3 = constraints(C3),
	portray_clause(Term3),

	see(File1),
	% writes the remaining clauses from the first file
	repeat,
		read_term(Rest1,[syntax_errors(dec10)]),
		(subsumes_chk((:- op(_,_,_)),Rest1) -> Rest1 = (:- Op1), call(Op1) ; write_file(Rest1)),
		Rest1 == end_of_file,
		!,
	seen,

	see(File2),
	% writes the remaining clauses from the second file
	repeat,
		read_term(Rest2,[syntax_errors(dec10)]),
		(subsumes_chk((:- op(_,_,_)),Rest2) -> Rest2 = (:- Op2), call(Op2) ; write_file(Rest2)),
		Rest2 == end_of_file,
		!,
	seen,

	told,
	tell(OldOutput),
	see(OldInput).

put_handler(File,Fileh) :-
	seeing(OldInput),
	telling(OldOutput),
	tell(Fileh),

	see(File),
	% writes 'use_module' clauses from the file (if any)

	repeat,
		read_term(Term,[syntax_errors(dec10)]),
		((subsumes_chk((:- use_module(_)),Term);subsumes_chk((:- use_module(_,_)),Term);
		subsumes_chk((:- demo_use_module(_)),Term);subsumes_chk((:- demo_use_module(_,_)),Term))
			-> write_file(Term) ; true),
		\+ (subsumes_chk((:- use_module(_)),Term);subsumes_chk((:- use_module(_,_)),Term);
		subsumes_chk((:- demo_use_module(_)),Term);subsumes_chk((:- demo_use_module(_,_)),Term)),
		!,

	portray_clause((handler demo_handler)),	%%% insert handler
	portray_clause(Term),	%%% constraints

	% writes the remaining clauses from the file
	repeat,
		read_term(Rest,[syntax_errors(dec10)]),
		write_file(Rest),
		Rest == end_of_file,
		!,
	seen,
	told,
	tell(OldOutput),
	see(OldInput).


write_file(end_of_file) :- !.
write_file(X) :- portray_clause(X), nl.

warn_singleton(File) :-
	seeing(OldInput),
	see(File),
	repeat,
		read_term(Rest,[syntax_errors(dec10),singletons(S),layout(L)]),
		( subsumes_chk([(Name=_)|_],S) -> [(Name=_)|_]=S, print_message(warning,demo(singleton(File,Name,L)))
		; true ),
		Rest == end_of_file,
		!,
	seen,
	see(OldInput).

%%% flattens a list
flatten([],[]) :- !.
flatten(X,[X]) :- \+ is_list(X).
flatten([H|T],F) :- flatten(H,FH), flatten(T,FT), append(FH,FT,F).

% Repeats the Command N times
x(Command, N) :-
	N>1 -> call(Command), M is N-1, x(Command, M) ;
	N==1 -> call(Command).
	
%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
